"""
Constants that will be required for the models
"""

GRAV = 6.67408e-8                 # Gravitational constant in CGS
BOLTZMANN = 1.38064852e-16            # Boltzmann
H1MASS = 1.6735328363560003e-24    # Mass of a Hydrogen atom in grams
PARSEC = 3.08567758123e18      # 1 parsec in cm
SOLARMASS = 1.989e33           # Solar mass in grams
YEAR = 60 * 60 * 24 * 365      # 1 year in seconds
