# It is with great thanks I mention James Tocknell (Github: Aragilar) for his continued assistance in providing
# solutions to my coding woes
"""
Defines common data structures and how they are to be written to files
"""
from ._containers import InitialConditions, ConfigParams, Solution, InternalData, SolutionPhysicalUnits
from ._dumpers import create_results_file
from ._loaders import solution_loader
